@@ -17,7 +17,7 @@ import shlex
 import urwid
 
 from . import Command, registerCommand
-from . import CommandCanceled
+from . import CommandCanceled, SequenceCanceled
 from .utils import update_keys
 from .. import commands
 
@@ -25,10 +25,10 @@ from .. import buffers
 from .. import helper
 from ..helper import split_commandstring
 from ..helper import mailto_to_envelope
-from ..completion import CommandLineCompleter
-from ..completion import ContactsCompleter
-from ..completion import AccountCompleter
-from ..completion import TagsCompleter
+from ..completion.commandline import CommandLineCompleter
+from ..completion.contacts import ContactsCompleter
+from ..completion.accounts import AccountCompleter
+from ..completion.tags import TagsCompleter
 from ..widgets.utils import DialogBox
 from ..db.errors import DatabaseLockedError
 from ..db.envelope import Envelope
@@ -280,10 +280,10 @@ class ExternalCommand(Command):
                 ret = str(e)
             else:
                 _, err = await proc.communicate(stdin.read() if stdin else None)
-            if proc.returncode == 0:
-                ret = 'success'
-            elif err:
-                ret = err.decode(urwid.util.detected_encoding)
+                if proc.returncode == 0:
+                    ret = 'success'
+                elif err:
+                    ret = err.decode(urwid.util.detected_encoding)
         else:
             with ui.paused():
                 try:
@@ -304,7 +304,10 @@ class ExternalCommand(Command):
             if self.on_success is not None:
                 self.on_success()
         else:
-            ui.notify(ret, priority='error')
+            msg = "editor has exited with error code {} -- {}".format(
+                    proc.returncode,
+                    ret or "No stderr output")
+            ui.notify(msg, priority='error')
         if self.refocus and callerbuffer in ui.buffers:
             logging.info('refocussing')
             ui.buffer_focus(callerbuffer)
@@ -337,12 +340,13 @@ class EditCommand(ExternalCommand):
         logging.debug('using editor_cmd: %s', editor_cmdstring)
 
         self.cmdlist = None
-        if '%s' in editor_cmdstring:
-            cmdstring = editor_cmdstring.replace('%s',
-                                                 helper.shell_quote(path))
-            self.cmdlist = split_commandstring(cmdstring)
-        else:
-            self.cmdlist = split_commandstring(editor_cmdstring) + [path]
+        if editor_cmdstring:
+            if '%s' in editor_cmdstring:
+                cmdstring = editor_cmdstring.replace('%s',
+                                                     helper.shell_quote(path))
+                self.cmdlist = split_commandstring(cmdstring)
+            else:
+                self.cmdlist = split_commandstring(editor_cmdstring) + [path]
 
         logging.debug({'spawn: ': self.spawn, 'in_thread': self.thread})
         ExternalCommand.__init__(self, self.cmdlist,
@@ -723,8 +727,8 @@ class ComposeCommand(Command):
     """compose a new email"""
     def __init__(
             self,
-            envelope=None, headers=None, template=None, sender=u'',
-            tags=None, subject=u'', to=None, cc=None, bcc=None, attach=None,
+            envelope=None, headers=None, template=None, sender='',
+            tags=None, subject='', to=None, cc=None, bcc=None, attach=None,
             omit_signature=False, spawn=None, rest=None, encrypt=False,
             **kwargs):
         """
@@ -779,58 +783,34 @@ class ComposeCommand(Command):
         self.encrypt = encrypt
         self.tags = tags
 
-    async def apply(self, ui):
-        if self.envelope is None:
-            if self.rest:
-                if self.rest.startswith('mailto'):
-                    self.envelope = mailto_to_envelope(self.rest)
-                else:
-                    self.envelope = Envelope()
-                    self.envelope.add('To', self.rest)
-            else:
-                self.envelope = Envelope()
-        if self.template is not None:
-            # get location of tempsdir, containing msg templates
-            tempdir = settings.get('template_dir')
+    class ApplyError(Exception):
+        pass
 
-            path = os.path.expanduser(self.template)
-            if not os.path.dirname(path):  # use tempsdir
-                if not os.path.isdir(tempdir):
-                    ui.notify('no templates directory: %s' % tempdir,
-                              priority='error')
-                    return
-                path = os.path.join(tempdir, path)
+    def _get_template(self, ui):
+        # get location of tempsdir, containing msg templates
+        tempdir = settings.get('template_dir')
 
-            if not os.path.isfile(path):
-                ui.notify('could not find template: %s' % path,
+        path = os.path.expanduser(self.template)
+        if not os.path.dirname(path):  # use tempsdir
+            if not os.path.isdir(tempdir):
+                ui.notify('no templates directory: %s' % tempdir,
                           priority='error')
-                return
-            try:
-                with open(path, 'rb') as f:
-                    template = helper.try_decode(f.read())
-                self.envelope.parse_template(template)
-            except Exception as e:
-                ui.notify(str(e), priority='error')
-                return
+                raise self.ApplyError()
+            path = os.path.join(tempdir, path)
 
-        # set forced headers
-        for key, value in self.headers.items():
-            self.envelope.add(key, value)
-
-        # set forced headers for separate parameters
-        if self.sender:
-            self.envelope.add('From', self.sender)
-        if self.subject:
-            self.envelope.add('Subject', self.subject)
-        if self.to:
-            self.envelope.add('To', ','.join(self.to))
-        if self.cc:
-            self.envelope.add('Cc', ','.join(self.cc))
-        if self.bcc:
-            self.envelope.add('Bcc', ','.join(self.bcc))
-        if self.tags:
-            self.envelope.tags = [t for t in self.tags.split(',') if t]
+        if not os.path.isfile(path):
+            ui.notify('could not find template: %s' % path,
+                      priority='error')
+            raise self.ApplyError()
+        try:
+            with open(path, 'rb') as f:
+                template = helper.try_decode(f.read())
+            self.envelope.parse_template(template)
+        except Exception as e:
+            ui.notify(str(e), priority='error')
+            raise self.ApplyError()
 
+    async def _get_sender_details(self, ui):
         # find out the right account, if possible yet
         account = self.envelope.account
         if account is None:
@@ -856,6 +836,8 @@ class ComposeCommand(Command):
 
                 ui.senderhistory.append(fromaddress)
                 self.envelope.add('From', fromaddress)
+        else:
+            fromaddress = self.envelope.get("From")
 
         # try to find the account again
         if account is None:
@@ -870,7 +852,8 @@ class ComposeCommand(Command):
         if self.envelope.account is None:
             self.envelope.account = account
 
-        # add signature
+    async def _set_signature(self, ui):
+        account = self.envelope.account
         if not self.omit_signature and account.signature:
             logging.debug('has signature')
             sig = os.path.expanduser(account.signature)
@@ -886,16 +869,54 @@ class ComposeCommand(Command):
                     mimetype = helper.guess_mimetype(sigcontent)
                     if mimetype.startswith('text'):
                         sigcontent = helper.try_decode(sigcontent)
-                        self.envelope.body += '\n' + sigcontent
+                        self.envelope.body_txt += '\n' + sigcontent
             else:
                 ui.notify('could not locate signature: %s' % sig,
                           priority='error')
                 if (await ui.choice('send without signature?', 'yes',
                                     'no')) == 'no':
-                    return
+                    raise self.ApplyError
 
-        # Figure out whether we should GPG sign messages by default
-        # and look up key if so
+    async def apply(self, ui):
+        try:
+            await self.__apply(ui)
+        except self.ApplyError:
+            return
+
+    def _get_account(self, ui):
+        # find out the right account
+        sender = self.envelope.get('From')
+        _, addr = email.utils.parseaddr(sender)
+        try:
+            account = settings.get_account_by_address(addr)
+        except NoMatchingAccount:
+            msg = 'Cannot compose mail - no account found for `%s`' % addr
+            logging.error(msg)
+            ui.notify(msg, priority='error')
+            raise CommandCanceled()
+
+        if account is None:
+            accounts = settings.get_accounts()
+            if not accounts:
+                ui.notify('no accounts set.', priority='error')
+                raise self.ApplyError
+            account = accounts[0]
+
+        return account
+
+    def _set_envelope(self):
+        if self.envelope is None:
+            if self.rest:
+                if self.rest.startswith('mailto'):
+                    self.envelope = mailto_to_envelope(self.rest)
+                else:
+                    self.envelope = Envelope()
+                    self.envelope.add('To', self.rest)
+            else:
+                self.envelope = Envelope()
+
+    def _set_gpg_sign(self, ui):
+        account = self.envelope.account
         if account.sign_by_default:
             if account.gpg_key:
                 self.envelope.sign = account.sign_by_default
@@ -906,7 +927,8 @@ class ComposeCommand(Command):
                 logging.warning(msg)
                 ui.notify(msg, priority='error')
 
-        # get missing To header
+    async def _set_to(self, ui):
+        account = self.envelope.account
         if 'To' not in self.envelope.headers:
             allbooks = not settings.get('complete_matching_abook_only')
             logging.debug(allbooks)
@@ -923,6 +945,43 @@ class ComposeCommand(Command):
             ui.recipienthistory.append(to)
             self.envelope.add('To', to)
 
+    async def _set_gpg_encrypt(self, ui):
+        account = self.envelope.account
+        if self.encrypt or account.encrypt_by_default == "all":
+            logging.debug("Trying to encrypt message because encrypt=%s and "
+                          "encrypt_by_default=%s", self.encrypt,
+                          account.encrypt_by_default)
+            await update_keys(ui, self.envelope, block_error=self.encrypt)
+        elif account.encrypt_by_default == "trusted":
+            logging.debug("Trying to encrypt message because "
+                          "account.encrypt_by_default=%s",
+                          account.encrypt_by_default)
+            await update_keys(ui, self.envelope, block_error=self.encrypt,
+                              signed_only=True)
+        else:
+            logging.debug("No encryption by default, encrypt_by_default=%s",
+                          account.encrypt_by_default)
+
+    def _set_base_attributes(self):
+        # set forced headers
+        for key, value in self.headers.items():
+            self.envelope.add(key, value)
+
+        # set forced headers for separate parameters
+        if self.sender:
+            self.envelope.add('From', self.sender)
+        if self.subject:
+            self.envelope.add('Subject', self.subject)
+        if self.to:
+            self.envelope.add('To', ','.join(self.to))
+        if self.cc:
+            self.envelope.add('Cc', ','.join(self.cc))
+        if self.bcc:
+            self.envelope.add('Bcc', ','.join(self.bcc))
+        if self.tags:
+            self.envelope.tags = [t for t in self.tags.split(',') if t]
+
+    async def _set_subject(self, ui):
         if settings.get('ask_subject') and \
                 'Subject' not in self.envelope.headers:
             subject = await ui.prompt('Subject')
@@ -932,6 +991,7 @@ class ComposeCommand(Command):
 
             self.envelope.add('Subject', subject)
 
+    async def _set_compose_tags(self, ui):
         if settings.get('compose_ask_tags'):
             comp = TagsCompleter(ui.dbman)
             tags = ','.join(self.tags) if self.tags else ''
@@ -942,27 +1002,37 @@ class ComposeCommand(Command):
 
             self.envelope.tags = tags
 
+    def _set_attachments(self):
         if self.attach:
             for gpath in self.attach:
                 for a in glob.glob(gpath):
                     self.envelope.attach(a)
                     logging.debug('attaching: %s', a)
 
+    async def __apply(self, ui):
+        self._set_envelope()
+        if self.template is not None:
+            self._get_template(ui)
+        # Set headers and tags
+        self._set_base_attributes()
+        # set account and missing From header
+        await self._get_sender_details(ui)
+
+        # add signature
+        await self._set_signature(ui)
+        # Figure out whether we should GPG sign messages by default
+        # and look up key if so
+        self._set_gpg_sign(ui)
+        # get missing To header
+        await self._set_to(ui)
+        # Set subject
+        await self._set_subject(ui)
+        # Add additional tags
+        await self._set_compose_tags(ui)
+        # Set attachments
+        self._set_attachments()
         # set encryption if needed
-        if self.encrypt or account.encrypt_by_default == u"all":
-            logging.debug("Trying to encrypt message because encrypt=%s and "
-                          "encrypt_by_default=%s", self.encrypt,
-                          account.encrypt_by_default)
-            await update_keys(ui, self.envelope, block_error=self.encrypt)
-        elif account.encrypt_by_default == u"trusted":
-            logging.debug("Trying to encrypt message because "
-                          "account.encrypt_by_default=%s",
-                          account.encrypt_by_default)
-            await update_keys(ui, self.envelope, block_error=self.encrypt,
-                              signed_only=True)
-        else:
-            logging.debug("No encryption by default, encrypt_by_default=%s",
-                          account.encrypt_by_default)
+        await self._set_gpg_encrypt(ui)
 
         cmd = commands.envelope.EditCommand(envelope=self.envelope,
                                             spawn=self.force_spawn,
@@ -970,6 +1040,7 @@ class ComposeCommand(Command):
         await ui.apply_command(cmd)
 
 
+
 @registerCommand(
     MODE, 'move', help='move focus in current buffer',
     arguments=[
@@ -1107,3 +1178,30 @@ class RemoveQueryCommand(Command):
         # flush index
         if self.flush:
             ui.apply_command(commands.globals.FlushCommand())
+
+
+@registerCommand(
+    MODE, 'confirmsequence',
+    arguments=[
+        (['msg'], {'help': 'Additional message to prompt',
+                   'nargs': '*'})
+    ],
+    help="prompt to confirm a sequence of commands")
+class ConfirmCommand(Command):
+    """Prompt user to confirm a sequence of commands."""
+
+    def __init__(self, msg=None, **kwargs):
+        """
+        :param msg: Additional message to prompt the user with
+        :type msg: List[str]
+        """
+        super(ConfirmCommand, self).__init__(**kwargs)
+        if not msg:
+            self.msg = "Confirm sequence?"
+        else:
+            self.msg = "Confirm sequence: {}?".format(" ".join(msg))
+
+    async def apply(self, ui):
+        if (await ui.choice(self.msg, select='yes', cancel='no',
+                            msg_position='left')) == 'no':
+            raise SequenceCanceled()
